"""Verify type hints behave corectly.

This doesn't have the test_ prefix, since runtime testing isn't particularly useful.
"""
from collections.abc import AsyncGenerator, Generator
from typing import List, NoReturn, Union

from typing_extensions import assert_never, assert_type

import outcome
from outcome import Error, Maybe, Outcome, Value, acapture, capture


class Super:
    """A superclass, inheriting ultimately from object."""


class Sub(Super):
    """A subclass."""


maybe: Maybe[float] = capture(len, [])
assert_type(maybe, Union[Value[float], Error])

# Check that this is immutable.
outcome.__version__ = 'dev'  # type: ignore[misc]


def maybe_test_val_first(maybe: Maybe[float]) -> None:
    """Check behaviour of the Maybe annotation, when checking Value first."""
    assert_type(maybe, Union[Value[float], Error])
    # Check narrowing.
    if isinstance(maybe, Value):
        assert_type(maybe, Value[float])
        assert_type(maybe.value, float)
        maybe.error  # type: ignore[attr-defined]
    else:
        assert_type(maybe, Error)
        if isinstance(maybe, Error):
            assert_type(maybe, Error)
            assert_type(maybe.error, BaseException)
            maybe.value  # type: ignore[attr-defined]
        else:
            assert_never(maybe)


def maybe_test_err_first(maybe: Maybe[float]) -> None:
    """Check behaviour of the Maybe annotation, when checking Error first."""
    assert_type(maybe, Union[Value[float], Error])
    # Check narrowing.
    if isinstance(maybe, Error):
        assert_type(maybe, Error)
        assert_type(maybe.error, BaseException)
        maybe.value  # type: ignore[attr-defined]
    else:
        assert_type(maybe, Value[float])
        if isinstance(maybe, Value):
            assert_type(maybe, Value[float])
            assert_type(maybe.value, float)
            maybe.error  # type: ignore[attr-defined]
        else:
            assert_never(maybe)


def value_variance_test() -> None:
    """Check variance behaves as expected."""
    value: Value[Super]
    value_sub: Value[Sub] = Value(Sub())
    value_super: Value[object] = Value(None)
    value = value_sub  # Is covariant.
    value = value_super  # type: ignore[assignment]


def outcome_test() -> None:
    """Test assigning either type to the base class."""
    value: Value[List[str]] = Value(['a', 'b', 'c'])
    bad_value: Value[float] = Value(48.3)
    error: Error = Error(Exception())

    outcome_good: Outcome[List[str]] = value
    outcome_mismatch: Outcome[bool] = value  # type: ignore[assignment]
    outcome_err: Outcome[List[str]] = error

    assert_type(outcome_good.unwrap(), List[str])
    assert_type(outcome_err.unwrap(), List[str])
    assert_type(value.unwrap(), List[str])
    assert_type(error.unwrap(), NoReturn)


def sync_raises() -> NoReturn:
    raise NotImplementedError


def sync_generator_one() -> Generator[int, str, List[str]]:
    word: str = (yield 5)
    assert len(word) == 3
    return ['a', 'b', 'c']


def sync_generator_two() -> Generator[str, int, List[str]]:
    three: int = (yield 'house')
    assert three.bit_length() == 2
    return ['a', 'b', 'c']


def sync_none() -> bool:
    return True


def sync_one(param: float) -> int:
    return round(param)


def sync_capture_test() -> None:
    """Test synchronous behaviour."""
    assert_type(capture(sync_none), Union[Value[bool], Error])
    assert_type(capture(sync_one, 3.14), Union[Value[int], Error])
    assert_type(capture(sync_one, param=3.14), Union[Value[int], Error])
    assert_type(capture(sync_raises), Error)
    capture(sync_one)  # type: ignore[arg-type, call-arg]
    capture(sync_none, 1, 2)  # type: ignore[arg-type, call-arg]


async def sync_gen_test() -> None:
    """Check send methods."""
    value_one: Value[str] = Value('abc')
    value_two: Value[int] = Value(3)
    error: Error = Error(Exception())
    outcome_one: Outcome[str] = [error, value_one][1]
    outcome_two: Outcome[int] = [error, value_two][1]

    assert_type(outcome_one.send(sync_generator_one()), int)
    assert_type(outcome_two.send(sync_generator_two()), str)
    outcome_one.send(sync_generator_two())  # type: ignore[arg-type]
    outcome_two.send(sync_generator_one())  # type: ignore[arg-type]

    assert_type(value_one.send(sync_generator_one()), int)
    assert_type(value_two.send(sync_generator_two()), str)
    value_one.send(sync_generator_two())  # type: ignore[arg-type]
    value_two.send(sync_generator_one())  # type: ignore[arg-type]

    # Error doesn't care.
    assert_type(error.send(sync_generator_one()), int)
    assert_type(error.send(sync_generator_two()), str)


async def async_none() -> bool:
    return True


async def async_raises() -> NoReturn:
    raise NotImplementedError


async def async_one(param: float) -> int:
    return round(param)


async def async_generator_one() -> AsyncGenerator[int, str]:
    assert_type((yield 5), str)


async def async_generator_two() -> AsyncGenerator[str, int]:
    assert_type((yield ''), int)


async def async_capture_test() -> None:
    """Test asynchronous behaviour."""
    assert_type(await acapture(async_none), Union[Value[bool], Error])
    assert_type(await acapture(async_one, 3.14), Union[Value[int], Error])
    assert_type(
        await acapture(async_one, param=3.14), Union[Value[int], Error]
    )
    assert_type(await acapture(async_raises), Error)
    capture(async_one)  # type: ignore[arg-type, call-arg]
    capture(async_none, 1, 2)  # type: ignore[arg-type, call-arg]


async def async_gen_test() -> None:
    value_one: Value[str] = Value('abc')
    value_two: Value[int] = Value(3)
    error: Error = Error(Exception())
    outcome_one: Outcome[str] = [error, value_one][1]
    outcome_two: Outcome[int] = [error, value_two][1]

    assert_type(await outcome_one.asend(async_generator_one()), int)
    assert_type(await outcome_two.asend(async_generator_two()), str)
    await outcome_one.asend(async_generator_two())  # type: ignore[arg-type]
    await outcome_two.asend(async_generator_one())  # type: ignore[arg-type]

    assert_type(await value_one.asend(async_generator_one()), int)
    assert_type(await value_two.asend(async_generator_two()), str)
    await value_one.asend(async_generator_two())  # type: ignore[arg-type]
    await value_two.asend(async_generator_one())  # type: ignore[arg-type]

    # Error doesn't care.
    assert_type(await error.asend(async_generator_one()), int)
    assert_type(await error.asend(async_generator_two()), str)



#!/usr/bin/env python3
#
# Documentation build configuration file, created by
# sphinx-quickstart on Sat Jan 21 19:11:14 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

# So autodoc can import our package
sys.path.insert(0, os.path.abspath('../..'))

# Warn about all references to unknown targets
nitpicky = True
# Except for these ones, which we expect to point to unknown targets:
nitpick_ignore = [
    # Format is ('sphinx reference type', 'string'), e.g.:
    ('py:obj', 'bytes-like'),
    # Typevars aren't found, for some reason.
    ('py:class', 'ArgsT'),
    ('py:class', 'ArgsT.args'),
    ('py:class', 'ArgsT.kwargs'),
    ('py:class', 'ResultT'),
    ('py:class', 'outcome._impl.ResultT'),
    ('py:class', 'outcome._impl.ValueT'),
]

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.coverage',
    'sphinx.ext.napoleon',
    'sphinxcontrib_trio',
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'trio': ('https://trio.readthedocs.io/en/stable', None),
}

autodoc_member_order = 'bysource'

# Add any paths that contain templates here, relative to this directory.
templates_path = []

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'outcome'
copyright = 'The outcome authors'
author = 'The outcome authors'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
import outcome

version = outcome.__version__
# The full version, including alpha/beta/rc tags.
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set 'language' from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# The default language for :: blocks
highlight_language = 'python3'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
#html_theme = 'alabaster'

# We have to set this ourselves, not only because it's useful for local
# testing, but also because if we don't then RTD will throw away our
# html_theme_options.
import sphinx_rtd_theme

html_theme = 'sphinx_rtd_theme'
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    # default is 2
    # show deeper nesting in the RTD theme's sidebar TOC
    # https://stackoverflow.com/questions/27669376/
    # I'm not 100% sure this actually does anything with our current
    # versions/settings...
    'navigation_depth': 4,
    'logo_only': True,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named 'default.css' will overwrite the builtin 'default.css'.
html_static_path = ['_static']

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'outcomedoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'outcome.tex', 'Trio Documentation', author, 'manual'),
]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, 'outcome', 'outcome Documentation', [author], 1)]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc, 'outcome', 'outcome Documentation', author, 'outcome',
        'Capture the outcome of Python function call.', 'Miscellaneous'
    ),
]
